/* evilwm - Minimalist Window Manager for X
 * Copyright (C) 1999-2002 Ciaran Anscomb <evilwm@6809.org.uk>
 * see README for license and other details. */

#include <fnmatch.h>
#include <stdlib.h>
#include <stdio.h>
#include <X11/extensions/shape.h>
#include "evilwm.h"

static int send_xmessage(Window w, Atom a, long x);

/* used all over the place.  return the client that has specified window as
 * either window or parent */

Client *
find_client(Window w)
{
	Client *c;

	for (c = head_client; c; c = c->next)
		if (w == c->parent || w == c->window)
			return c;
	return NULL;
}

Client *
find_client_by_name(const char *pat)
{
	Client *c;
	char *name;
	Window w = -1;
#ifdef DEBUG
	fprintf(stderr, "find by name: '%s'\n", pat);
#endif
	if (pat[0] == '0' && pat[1] == 'x')
		w = (Window)strtol(pat, NULL, 16);

	for (c = head_client; c != NULL; c = c->next) {
		if (c->window == w)
			return c;
		XFetchName(dpy, c->window, &name);
		if (name != NULL && fnmatch(pat, name, 0) == 0) {
			return c;
		}
	}
	return NULL;
}
				

/* iconify window and set state */
void
set_wm_state(Client *c, int state)
{
	long data[2];

	data[0] = (long) state;
	data[1] = None;

	XChangeProperty(dpy, c->window, xa_wm_state, xa_wm_state,
		32, PropModeReplace, (unsigned char *)data, 2);
}


int
wm_state(Client *c)
{
	Atom real_type;
	int real_format;
	unsigned long n, extra;
	unsigned char *data;
	long state = WithdrawnState;

	if ((XGetWindowProperty(dpy, c->window, xa_wm_state, 0L, 2L, False,
			AnyPropertyType, &real_type, &real_format, &n,
			&extra, &data) == Success) && n) {
		state = *(long *)data;
	}
	return state;
}

void
send_config(Client *c)
{
	XConfigureEvent ce;

	ce.type = ConfigureNotify;
	ce.event = c->window;
	ce.window = c->window;
	ce.x = c->x;
	ce.y = c->y;
	ce.width = c->width;
	ce.height = c->height;
	ce.border_width = 0;
	ce.above = None;
	ce.override_redirect = 0;

	XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
}

void
remove_client(Client *c)
{
	Client *p;

#ifdef DEBUG
	fprintf(stderr, "remove_client() : Removing...\n");
#endif

	XGrabServer(dpy);
	XSetErrorHandler(ignore_xerror);

	if (!quitting) {
#ifdef DEBUG
		fprintf(stderr, "\tremove_client() : setting WithdrawnState\n");
#endif
		set_wm_state(c, WithdrawnState);
		XRemoveFromSaveSet(dpy, c->window);
	}

	ungravitate(c);
	XSetWindowBorderWidth(dpy, c->window, 1);
	XReparentWindow(dpy, c->window, c->screen->root, c->x, c->y);
	if (c->parent)
		XDestroyWindow(dpy, c->parent);

	if (head_client == c) head_client = c->next;
	else for (p = head_client; p && p->next; p = p->next)
		if (p->next == c) p->next = c->next;

	if (c->size) XFree(c->size);
	if (current == c)
		current = NULL;  /* an enter event should set this up again */
	free(c);
#ifdef DEBUG
	{
		Client *i;
		int n = 0;
		for (i = head_client; i != NULL; i = i->next)
			n++;
		fprintf(stderr, "\tremove_client() : free(), window count now %d\n", n);
	}
#endif

	XSync(dpy, False);
	XSetErrorHandler(handle_xerror);
	XUngrabServer(dpy);
}

void
change_gravity(Client *c, int multiplier)
{
	int dx = 0, dy = 0;
	int gravity = (c->size->flags & PWinGravity) ?
		c->size->win_gravity : NorthWestGravity;

	switch (gravity) {
		case NorthWestGravity:
		case SouthWestGravity:
		case NorthEastGravity:
		case StaticGravity:
			dx = c->border;
		case NorthGravity:
			dy = c->border; break;
	}

	c->x += multiplier * dx;
	c->y += multiplier * dy;
#ifdef DEBUG
	if (dx || dy) {
		fprintf(stderr, "change_gravity() : window adjustment of %d,%d for ", multiplier * dx, multiplier * dy);
	switch (gravity) {
		case NorthWestGravity: fprintf(stderr, "NorthWestGravity\n"); break;
		case SouthWestGravity: fprintf(stderr, "SouthWestGravity\n"); break;
		case NorthEastGravity: fprintf(stderr, "NorthEastGravity\n"); break;
		case NorthGravity: fprintf(stderr, "NorthGravity\n"); break;
		case StaticGravity: fprintf(stderr, "StaticGravity\n"); break;
		default: fprintf(stderr, "unhandled gravity %d\n", gravity); break;
	}
	}
#endif
}

void
send_wm_delete(Client *c)
{
	int i, n, found = 0;
	Atom *protocols;

	if (c) {
		if (XGetWMProtocols(dpy, c->window, &protocols, &n)) {
			for (i=0; i<n; i++) if (protocols[i] == xa_wm_delete) found++;
			XFree(protocols);
		}
		if (found) send_xmessage(c->window, xa_wm_protos, xa_wm_delete);
		else XKillClient(dpy, c->window);
	}
}

static int
send_xmessage(Window w, Atom a, long x)
{
	XEvent ev;

	ev.type = ClientMessage;
	ev.xclient.window = w;
	ev.xclient.message_type = a;
	ev.xclient.format = 32;
	ev.xclient.data.l[0] = x;
	ev.xclient.data.l[1] = CurrentTime;

	return XSendEvent(dpy, w, False, NoEventMask, &ev);
}
/*
void client_to_front(Client *c) {
	Client *p;

	if (head_client == c) {
		head_client = c->next;
	}
	for (p = head_client; p->next; p = p->next) {
		if (p->next == c) {
			p->next = c->next;
		}
	}
	p->next = c;
	c->next = NULL;
}
*/

void
set_shape(Client *c)
{
	int n, order;
	XRectangle *rect;

	rect = XShapeGetRectangles(dpy, c->window, ShapeBounding, &n, &order);
	if (n > 1)
		XShapeCombineShape(dpy, c->parent, ShapeBounding, c->border,
				c->border, c->window, ShapeBounding, ShapeSet);
	XFree((void*)rect);
}

void
client_update_current(Client *c, Client *newcurrent)
{
	if (c) {
		if (c->vdesk == STICKY)
			XSetWindowBackground(dpy, c->parent, c == newcurrent ? c->screen->fc.pixel : c->screen->bg.pixel);
		else
			XSetWindowBackground(dpy, c->parent, c == newcurrent ? c->screen->fg.pixel : c->screen->bg.pixel);
		XClearWindow(dpy, c->parent);
	}
	current = newcurrent;
}
