/* evilwm - Minimalist Window Manager for X
 * Copyright (C) 1999-2002 Ciaran Anscomb <evilwm@6809.org.uk>
 * see README for license and other details. */

#include <X11/Xproto.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/extensions/xf86vmode.h>

#include "evilwm.h"

/* Now do this by fork()ing twice so we don't have to worry about SIGCHLDs */
void
spawn(const char *cmd[])
{
	pid_t pid;
#ifdef DEBUG
	int i;
#endif

	if (current_screen && current_screen->display)
		putenv(current_screen->display);
#ifdef DEBUG
	i = 0;
	while (cmd[i] != NULL) {
		fprintf(stderr, "EXEC %s\n", cmd[i++]);
	}
#endif
	if (!(pid = fork())) {
		setsid();
		switch (fork()) {
			/* explicitly hack around broken SUS execvp prototype */
			case 0: execvp(cmd[0], (char *const *)&cmd[1]);
			default: _exit(0);
		}
	}
	if (pid > 0)
		wait(NULL);
}

/**
 ** Signal handler called by Xt after a signal arrived - clean up and exit
 **
 ** (this function is also called directly to end the window manager)
 **/
void
handle_signal(XtPointer client_data, XtSignalId *signo)
{
	int i;
	/* SIGCHLD check no longer necessary */
	/* Quit Nicely */
	quitting = 1;
	while(head_client) remove_client(head_client);
	XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
	if (font) XFreeFont(dpy, font);
	for (i = 0; i < num_screens; i++) {
		XFreeGC(dpy, screens[i].invert_gc);
		XInstallColormap(dpy, DefaultColormap(dpy, screens[i].screen));
	}
	free(screens);
	XCloseDisplay(dpy);
	exit(0);
}

int
handle_xerror(Display *dsply, XErrorEvent *e)
{
	Client *c = find_client(e->resourceid);

	/* If this error actually occurred while setting up the new
	 * window, best let make_new_client() know not to bother */
	if (initialising != None && e->resourceid == initialising) {
#ifdef DEBUG
		fprintf(stderr, "\t **SAVED?** handle_xerror() caught error %d while initialising\n", e->error_code);
#endif
		initialising = None;
		return 0;
	}
#ifdef DEBUG
	fprintf(stderr, "**ERK** handle_xerror() caught an XErrorEvent: %d\n", e->error_code);
#endif
	/* if (e->error_code == BadAccess && e->resourceid == root) { */
	if (e->error_code == BadAccess && e->request_code == X_ChangeWindowAttributes) {
#ifdef STDIO
		fprintf(stderr, "root window unavailable (maybe another wm is running?)\n");
#endif
		exit(1);
	}
#ifdef XDEBUG
	fprintf(stderr, "XError %x ", e->error_code);
#endif
	/* Kludge around IE misbehaviour */
	if (e->error_code == 0x8 && e->request_code == 0x0c && e->minor_code == 0x00) {
#ifdef DEBUG
		fprintf(stderr, "\thandle_xerror() : IE kludge - ignoring XError\n");
#endif
		return 0;
	}

	if (c) {
#ifdef DEBUG
		fprintf(stderr, "\thandle_xerror() : calling remove_client()\n");
#endif
		remove_client(c);
	}
	return 0;
}

int
ignore_xerror(Display *dsply, XErrorEvent *e)
{
#ifdef DEBUG
	fprintf(stderr, "ignore_xerror() caught an XErrorEvent: %d\n", e->error_code);
#endif
	return 0;
}

void
dump_clients(FILE *f)
{
	Client *c;
	char *name;

	for (c = head_client; c; c = c->next) {
		XFetchName(dpy, c->window, &name);
		fprintf(f, "LIST %s 0x%x %s (%d, %d) @ %d,%d %dx%d\n",
		    c == current ? "*" : " ", (unsigned int)c->window, name,
		    wm_state(c), c->vdesk, c->x, c->y, c->width, c->height);

	}
}

int
set_gamma(Display *dpy, int screen, double gam)
{
	XF86VidModeGamma gamma;
	
	gamma.red = gam;
	gamma.green = gam;
	gamma.blue = gam;

	if (!XF86VidModeSetGamma(dpy, screen, &gamma))
		return -1;
	else
		return 0;
}

	    
